/****************************************************************************
	JTAG BounderyScan Access Library

	Copyright(c) 2003 Tatsuyuki Satoh

	note:

	issue:
		TDO readback data is not correct when daisy chained target.

	Histry
		2002. 5.29 change bit order MSB-FIRST to input data
		           added bit order conversion potion for driver I/F.
		2002. 3.11 separated from spcisp.c

****************************************************************************/

#include <windows.h>
#include <stdio.h>
#include "devlist.h"
#include "bscan.h"

/* JTAG Boundary Scan RAW Access Function */
#if JTAG_DYNAMIC_DRIVER
int (*jtag_txrx)(const BYTE *ibuf,BYTE *obuf,DWORD drsize,int tms_s,int tms_e);
#endif

/* for Driver Adjuster */
int bscan_convert_bit_order_in = 0; // convert bit order input data

/* for DaisyChain Adjuster */
int bscan_num_chain;		// number of daisy chain chips
int bscan_tdo_chips;		// number of chips in TDO side for target
int bscan_tdo_ir;			// number of total IR bits in back chain
int bscan_tdi_ir;			// number of total IR bits in TDI side chain
int bscan_tdi_bypass;		// number of chips in TDI side chain
int bscan_tdo_bypass;		// number of chips in TDO side chain
int bscan_before_dr;		// pre-shift  dummy shift before DR transfer
int bscan_after_dr;		// post-shift dummy shift after  DR transfer

/****************************************************************************/
/* swap bits in BYTE data                                                   */
/****************************************************************************/
static BYTE swap_byte(BYTE value)
{
	BYTE ret = 0;

	if(value&0x01) ret |= 0x80;
	if(value&0x02) ret |= 0x40;
	if(value&0x04) ret |= 0x20;
	if(value&0x08) ret |= 0x10;
	if(value&0x10) ret |= 0x08;
	if(value&0x20) ret |= 0x04;
	if(value&0x40) ret |= 0x02;
	if(value&0x80) ret |= 0x01;

	return ret;
}


/****************************************************************************/
/* BounderyScan reset(IDLE)                                                 */
/****************************************************************************/
int bscan_reset(int num_tck)
{
	/* Logic reset +  run idle mode */
	jtag_txrx(NULL,NULL,num_tck,1,0);

	return 0;
}

/****************************************************************************/
/* BounderyScan IDLE to IR and DR, and IDLE                                 */
/****************************************************************************/
int bscan_inst(BYTE cmd,const BYTE *ibuf,BYTE *obuf,int drsize,int finish)
{
	int i;
	int ret = 0;
	const char bypass_ir[8] =
{	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff };

	// Enter SELECT-IR state : 2clock
	jtag_txrx(NULL,NULL,2,1,1);

	// Enter SHIFT-IR state : 2clock
	jtag_txrx(NULL,NULL,2,0,0);

	// Shift BypassCommand to TDO side
	if(bscan_tdo_ir)
		jtag_txrx(bypass_ir,NULL,bscan_tdo_ir,0,0);

	// LSB First to MSB First
	cmd = swap_byte(cmd);

	if(bscan_tdi_ir)
	{
		// Shift Instruction Code
		jtag_txrx(&cmd,NULL,5,0,0);
		// Shift BypassCommand to TDI side & exiting SHIFT-IR
		jtag_txrx(bypass_ir,NULL,bscan_tdi_ir,0,1);
	}
	else
	{
		// Shift Instruction Code & exiting SHIFT-IR
		jtag_txrx(&cmd,NULL,5,0,1);
	}

	// Etner SELECT-DR state
	jtag_txrx(NULL,NULL,2,1,1);

	// Etner SHIFT-DR state
	jtag_txrx(NULL,NULL,2,0,0);

	// Shift TDI bypass chain
	if(bscan_before_dr)
		jtag_txrx(NULL,NULL,bscan_before_dr,0,0);
	/* !!!!! get OBUF when bscan_before_dr>bscan_after_dr) !!!!! */

	// continue to Shift DR
	if(drsize)
	{
		ret = bscan_dr(ibuf,obuf,drsize,finish);

		// Convert Bit order
		if(obuf && bscan_convert_bit_order_in)
		{
			for(i=0;i<(drsize+7)/8;i++)
				obuf[i] = swap_byte(obuf[i]);
		}
	}
	return ret;
}

/****************************************************************************/
/* BounderyScan DR shift phase to IDLE phase                                */
/****************************************************************************/
int bscan_dr(const BYTE *ibuf,BYTE *obuf,int drsize,int finish)
{
	int c;
	int i;

	if(finish)
	{
		if(bscan_after_dr)
		{
			/* shift TDI */
			jtag_txrx(ibuf,obuf,drsize,0,0);
			/* shift TDI bypass */
			jtag_txrx(NULL,NULL,bscan_after_dr,0,1);
		}
		else
		{
			/* shift TDI */
			jtag_txrx(ibuf,obuf,drsize,0,1);
		}
		/* !!!!! shift OBUF and store captured TDI) !!!!! */

		/* update -DR state */
		jtag_txrx(NULL,NULL,1,1,1);

		/* RUN test IDLE */
		jtag_txrx(NULL,NULL,2,0,0);
	}
	else
	{
		/* can't read when don't finish with daisy chain */
#if 0
		if(obuf && (bscan_before_dr || bscan_after_dr))
		{
			return -1;
		}
#endif
		jtag_txrx(ibuf,obuf,drsize,0,0);
	}

	// Convert Bit order
	if(obuf && bscan_convert_bit_order_in)
	{
		for(i=0;i<(drsize+7)/8;i++)
			obuf[i] = swap_byte(obuf[i]);
	}

	return 0;
}

/****************************************************************************/
/* search total chain chips & total IR bits                                 */
/****************************************************************************/
#define MAX_CHAIN 8
static DEVLIST dev_chain[MAX_CHAIN];

// search total chain chips & total IR bits
int bscan_scan_chain(void)
{
	BYTE irbuf[MAX_CHAIN];
	BYTE idbuf[MAX_CHAIN * 4];
	BYTE odata , idata;
	int i;
	int total_ir;
	int last_ir;

	/* RESET LOGIC & RUN IDLE */
	jtag_txrx(NULL,NULL,6,1,0);

	// Enter SELECT-IR state
	jtag_txrx(NULL,NULL,2,1,1);
	// Enter SHIFT-IR state
	jtag_txrx(NULL,NULL,2,0,0);

	// sample IR & clear IR
	jtag_txrx(NULL,irbuf,MAX_CHAIN*8,0,0);

	// shift IR with BYPASS command & get number of ir's
	odata = 0xff;
	for(total_ir=0; total_ir<MAX_CHAIN*8 ; total_ir++)
	{
		jtag_txrx(&odata,&idata,1,0,0);
		if(idata!=0x00)
			break;
	}

	// Update IR (BYPASS command)
	jtag_txrx(&odata,NULL,1,0,1);
	// Etner SELECT-DR state
	jtag_txrx(NULL,NULL,2,1,1);
	// Etner SHIFT-DR state
	jtag_txrx(NULL,NULL,2,0,0);

	// shift bypass register & get num of chips
	odata = 0xff;
	for(bscan_num_chain=0; ; bscan_num_chain++)
	{
		jtag_txrx(&odata,&idata,1,0,0);
		if(idata!=0x00)
			break;
	}
	//update -DR state
	jtag_txrx(NULL,NULL,1,1,1);
#if 0
	printf("number of JTAG-chain %d , number of total IR\n",num_chain,total_ir);*
#endif
	/* RESET LOGIC & RUN IDLE */
	jtag_txrx(NULL,NULL,6,1,0);
	// Enter SELECT-DR state
	jtag_txrx(NULL,NULL,1,1,1);
	// Etner SHIFT-DR state
	jtag_txrx(NULL,NULL,2,0,0);

	// sampling DR data
	jtag_txrx(NULL,idbuf,32*bscan_num_chain,0,1);

	// Convert Bit order
	if(bscan_convert_bit_order_in)
	{
		for(i=0;i<MAX_CHAIN;i++)
			irbuf[i] = swap_byte(irbuf[i]);
		for(i=0;i<4*bscan_num_chain;i++)
			idbuf[i] = swap_byte(idbuf[i]);
	}

	// analyze DaisyChainDevice
	if( build_device_chain(dev_chain,MAX_CHAIN,irbuf,total_ir,idbuf,bscan_num_chain*32) != bscan_num_chain)
		return 0;

	// OK
	return 1;
}

/****************************************************************************/
/* devide auto ditect                                                       */
/****************************************************************************/
int bscan_device_autoditect(DWORD deviceid)
{
	int target_id = 0;
	int i;

	for(i=0 ; i<bscan_num_chain ; i++)
	{
		if( (dev_chain[i].id&0x0fffffff) == deviceid )
		{
			if(target_id !=0)
				return -1; /* two same device found */
			target_id = i;
		}
	}
	return target_id;
}

/****************************************************************************/
/* devide select                                                            */
/****************************************************************************/
void bscan_select(int position)
{
	int i;

	// setup TDO side bypass
	bscan_after_dr  = bscan_tdo_bypass = bscan_tdo_chips;
	bscan_before_dr = bscan_tdi_bypass = bscan_num_chain-bscan_tdo_chips-1;
	bscan_tdo_ir = bscan_tdi_ir = 0;

	printf("JTAG Chain : TDO ");
	for(i=0 ; i<bscan_num_chain ; i++)
	{
		printf("-<-%c%s[%08XH]",i==bscan_tdo_chips ? '*':' ',dev_chain[i].name,dev_chain[i].id);
		if(i<bscan_tdo_chips)
			bscan_tdo_ir += dev_chain[i].ir;
		else if(i>bscan_tdo_chips)
			bscan_tdi_ir += dev_chain[i].ir;
	}
	printf("-<-TDI\n");
#if 0
	printf("IR %d:%d , DR %d:%d\n",bscan_tdo_ir,bscan_tdi_ir,bscan_tdo_bypass,bscan_tdi_bypass);
#endif
}

